from rakam.connection import RakamConnection


class RakamClient(object):
    """
        Client class used for calling endpoints on rakam server.
    """
    def __init__(self, rakam_url, rakam_credentials=None, connection_options=None):
        connection_kwargs = {}
        if rakam_credentials is not None:
            connection_kwargs['rakam_credentials'] = rakam_credentials

        if connection_options is not None:
            connection_kwargs['options'] = connection_options

        self.connection = RakamConnection(rakam_url, **connection_kwargs)

    def send_bulk_events(self, events):
        return self.connection.send_bulk_events(events)

    def commit_bulk_events(self, timeout=None, raise_on_read_timeout=False):
        return self.connection.commit_bulk_events(timeout=timeout, raise_on_read_timeout=raise_on_read_timeout)

    def set_user_properties(self, _id, properties=None, timeout=None, raise_on_read_timeout=False):
        return self.connection.set_user_properties(
            _id, properties=properties, timeout=timeout, raise_on_read_timeout=raise_on_read_timeout)

    def user_batch_operations(self, operations, timeout=None, raise_on_read_timeout=False):
        return self.connection.user_batch_operations(operations, timeout=timeout, raise_on_read_timeout=raise_on_read_timeout)

    def user_batch_update(self, users, timeout=None, raise_on_read_timeout=False):
        return self.connection.user_batch_update(users, timeout=timeout, raise_on_read_timeout=raise_on_read_timeout)

    def execute_sql(self, query, timeout=None):
        return self.connection.execute_sql(query, timeout=timeout)



import pytz
from itertools import izip
from collections import OrderedDict
from six import string_types

from rakam.utils import parse_date, parse_time, parse_datetime
from rakam.exceptions import BadSqlResponse, SqlRequestFailed, \
    BadSqlMetadata, BadSqlResult


def _parse_date(val):
    return parse_date(val)


def _parse_time(val):
    return parse_time(val)


def _parse_datetime(val):
    return parse_datetime(val).astimezone(pytz.utc)


class RakamFieldMeta(object):
    RAKAM_FIELD_SERIALIZERS = {
        'STRING': unicode,
        'INTEGER': int,
        'DECIMAL': float,
        'DOUBLE': float,
        'LONG': long,
        'BOOLEAN': bool,
        'DATE': _parse_date,
        'TIME': _parse_time,
        'TIMESTAMP': _parse_datetime,
        'ARRAY_STRING': lambda col: map(unicode, col),
        'ARRAY_INTEGER': lambda col: map(int, col),
        'ARRAY_DECIMAL': lambda col: map(float, col),
        'ARRAY_DOUBLE': lambda col: map(float, col),
        'ARRAY_LONG': lambda col: map(long, col),
        'ARRAY_BOOLEAN': lambda col: map(bool, col),
        'ARRAY_DATE': lambda col: map(_parse_date, col),
        'ARRAY_TIME': lambda col: map(_parse_time, col),
        'ARRAY_TIMESTAMP': lambda col: map(_parse_datetime, col),
    }

    def __init__(self, name, _type, unique=False, descriptive_name='', description=None, category=None):
        self.name = name
        self.type = _type
        self.unique = unique
        self.descriptive_name = descriptive_name
        self.description = description
        self.category = category

        self.serializer = self._get_serializer()

    def _get_serializer(self):
        return self.RAKAM_FIELD_SERIALIZERS.get(self.type, lambda col: col)


class RakamSqlResult(object):
    def __init__(self, rows, metadata, properties=None):
        if properties is None:
            properties = {}

        self.rows = rows
        self.metadata = metadata
        self.properties = properties


class RakamSqlParser(object):
    iterable_types = (list, tuple)
    dict_types = (dict, OrderedDict)

    def parse(self, sql_result):
        failed = sql_result.get('failed', False)
        properties = sql_result.get('properties', None)
        result = sql_result.get('result', None)
        metadata = sql_result.get('metadata', None)
        error = sql_result.get('error', None)

        if failed:
            raise SqlRequestFailed("sql query failed. error: %s" % (error,))

        if properties is None:
            raise BadSqlResponse("properties is missing.")

        if not isinstance(properties, self.dict_types):
            raise BadSqlResponse("properties is must be dict. given: %s" % (type(properties),))

        if result is None:
            raise BadSqlResponse("result is missing.")

        if not isinstance(result, self.iterable_types):
            raise BadSqlResponse("result must be a list. given: %s", (type(result),))

        if metadata is None:
            raise BadSqlResponse("metadata is missing.")

        if not isinstance(metadata, self.iterable_types):
            raise BadSqlResponse("metadata must be a list. given: %s", (type(metadata),))

        field_metas = self._parse_metadata(metadata)
        rows = self._parse_rows(result, field_metas)
        return RakamSqlResult(rows, field_metas, properties=properties)

    def _parse_metadata(self, metadata_list):
        return map(self._parse_meta, metadata_list)

    def _parse_meta(self, meta_dict):
        name = meta_dict.get('name', None)
        _type = meta_dict.get('type', None)
        unique = meta_dict.get('unique', False)
        descriptive_name = meta_dict.get('descriptiveName', '')
        description = meta_dict.get('description', None)
        category = meta_dict.get('category')

        if name is None:
            raise BadSqlMetadata("name is missing.")

        if not isinstance(name, string_types):
            raise BadSqlMetadata("name must be a string. given: %s" % (type(name),))

        if _type is None:
            raise BadSqlMetadata("type is missing.")

        if not isinstance(_type, string_types):
            raise BadSqlMetadata("type must be a string. given: %s" % (type(_type),))

        return RakamFieldMeta(
            name, _type, unique=unique,
            descriptive_name=descriptive_name,
            description=description,
            category=category
        )

    def _parse_rows(self, rows_raw, field_metas):
        num_fields = len(field_metas)
        row_types = self.iterable_types

        rows = []
        for row_index, row_raw in enumerate(rows_raw):
            if not isinstance(row_raw, row_types):
                raise BadSqlResult("Invalid type for row at index: %s" % (row_index,))

            if len(row_raw) != num_fields:
                raise BadSqlResult("Invalid number of fields for row at index %s: %s" % (row_index, len(row_raw),))

            row = self._parse_row(row_raw, field_metas, row_index)
            rows.append(row)

        return rows

    def _parse_row(self, row_raw, field_metas, row_index):
        parse_column = self._parse_column
        row = tuple(
            parse_column(col, field_meta, row_index, col_index)
            for col, field_meta, col_index
            in izip(row_raw, field_metas, xrange(len(row_raw)))
        )

        return row

    def _parse_column(self, col_raw, field_meta, row_index, col_index):
        if col_raw is not None:
            try:
                return field_meta.serializer(col_raw)
            except Exception as exc:
                raise BadSqlResult("Invalid value at row: %s column: %s error: %s" % (row_index, col_index, exc))



# -*- coding: utf-8 -*-
"""
    Rakam client for limited use.
    Project is only made for use in multiple internal projects, hence it only
    contains a small amount of functionality.
"""


__version__ = '0.1.5'
__author__ = 'Kaan Uğurlu'
__contact__ = 'kugurlu94@gmail.com'
__all__ = ['RakamClient', 'RakamCredentials']


from rakam.connection import RakamCredentials
from rakam.client import RakamClient




class RakamError(Exception):
    """
        Generic rakam error.
    """
    pass


class MissingKey(Exception):
    """
        Required key for the action is not provided.
    """
    pass


class InvalidKeyError(RakamError):
    """
        Key is used for other type of requests.
    """
    pass


class BadKeyError(RakamError):
    """
        Key doesn't belong to your project.
    """
    pass


class ConflictError(RakamError):
    """
        Raised when same action is already being processed
    """
    pass


class RakamSqlError(RakamError):
    pass


class BadSqlResponse(RakamSqlError):
    """
        Sql result body is invalid.
    """
    pass


class SqlRequestFailed(RakamSqlError):
    """
        If sql query fails.
    """
    pass


class BadSqlMetadata(RakamSqlError):
    """
        Invalid metadata.
    """
    pass


class BadSqlResult(RakamSqlError):
    """
        Invalid result.
    """
    pass



import json
import logging
from collections import OrderedDict

from requests import Session, RequestException, ReadTimeout

from rakam.exceptions import RakamError, MissingKey, InvalidKeyError, BadKeyError, ConflictError
from rakam.sql import RakamSqlParser


logger = logging.getLogger('rakam.connection')


_EMPTY_PROPERTIES = {}


class RakamCredentials(object):
    def __init__(self, read_key=None, write_key=None, master_key=None):
        self.read_key = read_key
        self.write_key = write_key
        self.master_key = master_key


class RakamConnection(object):
    """
        Communication with rakam server through a persistent connection.
    """

    bulk_uri = "/event/bulk"
    commit_uri = "/event/bulk/commit"
    query_uri = "/query/execute"
    set_user_properties_uri = "/user/set_properties"
    user_batch_operations_uri = "/user/batch_operations"

    def __init__(self, rakam_url, rakam_credentials=None, options=None):
        if options is None:
            options = {}

        self.rakam_url = rakam_url
        self.rakam_credentials = rakam_credentials or RakamCredentials()
        self.write_retry = options.get('write_retry', 0)
        self.write_timeout = options.get('write_timeout', None)
        self.session = Session()
        self.sql_parser = RakamSqlParser()

    def send_bulk_events(self, events):
        max_retries = max(self.write_retry or 0, 0)
        retries = 0

        url = self.rakam_url + self.bulk_uri
        api_key = self._get_master_key()
        timeout = self.write_timeout

        while True:
            try:
                response = self._send_post(
                    url,
                    params={'commit': 'false'},
                    data=json.dumps(
                        OrderedDict(
                            [
                                ('api', {'writeKey': api_key}),
                                (
                                    'events',
                                    [
                                        OrderedDict(
                                            (
                                                ("collection", event['collection']),
                                                ("properties", event.get('properties', _EMPTY_PROPERTIES))
                                            )
                                        )
                                        for event in events
                                    ]
                                )
                            ]
                        )
                    ),
                    headers={'Content-Type': 'application/json'},
                    timeout=timeout
                )
            except ReadTimeout:  # Read timeout means events might be added but response not received.
                raise RakamError("Read timeout on bulk.")
            except RequestException as request_exc:
                if retries < max_retries:
                    retries += 1
                    logger.info("Error at rakam request, retrying... Attempt: %s", retries)
                    # retries once more
                else:
                    raise RakamError("Sending events failed with: %s after %s retries" % (request_exc, retries))
            else:
                status_code = int(response.status_code)
                if status_code / 100 != 2:
                    if status_code / 100 == 5 and retries < max_retries:  # Server error, lets retry.
                        retries += 1
                        logger.info("Server error at rakam request, retrying... Attempt: %s", retries)
                        # retries once more
                    else:
                        if status_code == 401:
                            raise InvalidKeyError("Invalid rakam key. Please check that you are using the master key.")
                        elif status_code == 403:
                            self._raise_for_403(response)
                        else:
                            self._raise_for_unknown(response)

                else:
                    break  # break on sucess.

        return True

    def commit_bulk_events(self, timeout=None, raise_on_read_timeout=False):
        """
            Commits bulk events recently sent to api.
            Returns a True if commit is finished successfully.
            Returns False if read timeout occurs (Commit continues even though a timeout.)

            Raises a subclass of RakamError if request was unsuccessfull or a ConnectionTimeout occurs.
        """
        url = self.rakam_url + self.commit_uri
        api_key = self._get_master_key()

        request_kwargs = {
            'headers': {
                'master_key': api_key,
                'Content-Type': 'application/json',
            },
            'data': json.dumps({"collections": None})
        }

        if timeout is not None:
            request_kwargs['timeout'] = timeout

        success = False
        try:
            response = self._send_post(url, **request_kwargs)
        except ReadTimeout:  # Read timeout is assumed to be processing
            if raise_on_read_timeout:
                raise RakamError("Read timeout on commit.")
            else:
                logger.info("Read timeout on commit, silently passing...")

        except RequestException as request_exc:
            raise RakamError("Http request failed with: %s" % (request_exc,))
        else:
            status_code = int(response.status_code)
            if status_code / 100 == 2:
                success = True
            elif status_code == 401:
                raise InvalidKeyError("Invalid rakam key. Please check that you are using the master key.")
            elif status_code == 403:
                self._raise_for_403(response)
            elif status_code == 409:
                raise ConflictError("Another commit is being proccessed.")
            else:
                self._raise_for_unknown(response)

        return success

    def set_user_properties(self, _id, properties=None, timeout=None, raise_on_read_timeout=False):
        url = self.rakam_url + self.set_user_properties_uri
        api_key = self._get_write_key()

        if properties is None:
            properties = {}

        request_kwargs = {
            'headers': {
                'Content-Type': 'application/json',
            },
            'data': json.dumps(
                {
                    'id': _id,
                    'api': {
                        'api_key': api_key,
                    },
                    'properties': properties,
                }
            )
        }

        if timeout is not None:
            request_kwargs['timeout'] = timeout

        success = False
        try:
            response = self._send_post(url, **request_kwargs)
        except ReadTimeout:  # Read timeout is assumed to be processing
            if raise_on_read_timeout:
                raise RakamError("Read timeout on set_user_properties.")
            else:
                logger.info("Read timeout on set_user_properties, silently passing...")
        except RequestException as request_exc:
            raise RakamError("set_user_properties failed with: %s" % (request_exc,))
        else:
            status_code = int(response.status_code)
            if status_code / 100 == 2:
                success = True
            elif status_code == 401:
                raise InvalidKeyError("Invalid rakam key. Please check that you are using the write key.")
            elif status_code == 403:
                self._raise_for_403(response)
            else:
                self._raise_for_unknown(response)

        return success

    def user_batch_operations(self, operations, timeout=None, raise_on_read_timeout=False):
        url = self.rakam_url + self.user_batch_operations_uri
        api_key = self._get_master_key()

        request_kwargs = {
            'headers': {
                'Content-Type': 'application/json',
            },
            'data': json.dumps(
                {
                    'api': {
                        'api_key': api_key,
                    },
                    'data': operations,
                }
            ),
        }

        if timeout is not None:
            request_kwargs['timeout'] = timeout

        success = False
        try:
            response = self._send_post(url, **request_kwargs)
        except ReadTimeout:  # Read timeout is assumed to be processing
            if raise_on_read_timeout:
                raise RakamError("Read timeout on user_batch_operations.")
            else:
                logger.info("Read timeout on user_batch_operations, silently passing...")
        except RequestException as request_exc:
            raise RakamError("user_batch_operations failed with: %s" % (request_exc,))
        else:
            status_code = int(response.status_code)
            if status_code / 100 == 2:
                success = True
            elif status_code == 401:
                raise InvalidKeyError("Invalid rakam key. Please check that you are using the master key.")
            elif status_code == 403:
                self._raise_for_403(response)
            else:
                self._raise_for_unknown(response)

        return success

    def user_batch_update(self, users, timeout=None, raise_on_read_timeout=False):
        return self.user_batch_operations(
            [
                {
                    'id': user['id'],
                    'set_properties': user['properties'],
                } for user in users
            ],
            timeout=timeout, raise_on_read_timeout=raise_on_read_timeout
        )

    def execute_sql(self, query, timeout=None):
        url = self.rakam_url + self.query_uri
        api_key = self._get_read_key()

        request_kwargs = {
            'headers': {
                'read_key': api_key,
                'Content-Type': 'application/json',
            },
            'data': json.dumps(
                {
                    "query": query,
                    "export_type": "JSON",
                }
            ),
        }

        if timeout is not None:
            request_kwargs['timeout'] = timeout

        try:
            response = self._send_post(
                url,
                **request_kwargs
            )
        except RequestException as request_exc:
            raise RakamError("Http request failed with: %s" % (request_exc,))
        else:
            status_code = int(response.status_code)
            if status_code / 100 == 2:
                return self._parse_sql_response(response.json())
            elif status_code == 401:
                raise InvalidKeyError("Invalid rakam key. Please check that you are using the read key.")
            elif status_code == 403:
                self._raise_for_403(response)
            else:
                self._raise_for_unknown(response)

    def _parse_sql_response(self, sql_result):
        return self.sql_parser.parse(sql_result)

    def _parse_error_body(self, json_body):
        return json_body.get("error", None)

    def _get_read_key(self):
        read_key = self.rakam_credentials.read_key
        if read_key is not None:
            return read_key
        else:
            raise MissingKey("read_key is not provided.")

    def _get_write_key(self):
        write_key = self.rakam_credentials.write_key
        if write_key is not None:
            return write_key
        else:
            raise MissingKey("write_key is not provided.")

    def _get_master_key(self):
        master_key = self.rakam_credentials.master_key
        if master_key is not None:
            return master_key
        else:
            raise MissingKey("master_key is not provided.")

    # Request sending methods
    def _send_get(self, url, **kwargs):
        return self._send_request(url, method='GET', **kwargs)

    def _send_post(self, url, **kwargs):
        return self._send_request(url, method='POST', **kwargs)

    def _send_request(self, url, method='GET', data='', params=None, headers=None, timeout=None):
        request_kwargs = {'url': url, 'data': data}
        if params is not None:
            request_kwargs['params'] = params

        if headers is not None:
            request_kwargs['headers'] = headers

        if timeout is not None:
            request_kwargs['timeout'] = timeout

        method_upper = method.upper()
        if method_upper == 'GET':
            send_method = self.session.get
        elif method_upper == 'POST':
            send_method = self.session.post
        else:
            raise RakamError("%s method is not supported." % (method,))

        return send_method(**request_kwargs)

    # Response handling methods
    def _parse_error_response(self, response):
        try:
            message = self._parse_error_body(response.json())
        except ValueError:
            pass

        return message

    def _raise_for_403(self, response):
        message = self._parse_error_response(response)
        if message is None:
            message = "Wrong rakam key. Please check that you are using the key for project."

        raise BadKeyError(message)

    def _raise_for_unknown(self, response):
        status_code = response.status_code
        try:
            error_message = response.json()
        except ValueError:
            raise RakamError(
                "Request failed with status: %s body: %s" % (
                    status_code, response.text
                )
            )
        except:
            raise RakamError(
                "Request failed with status: %s" % (status_code,)
            )
        else:
            raise RakamError(
                "Request failed with status: %s error_code: %s error: %s" % (
                    status_code,
                    error_message.get('error_code', ''),
                    error_message.get('error', ''),
                )
            )



# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages

module_path = os.path.join(os.path.dirname(__file__), 'rakam/__init__.py')
version_line = [line for line in open(module_path)
                if line.startswith('__version__')][0]

__version__ = version_line.split('__version__ = ')[-1][1:][:-2]


setup(
    name="rakam",
    version=__version__,

    url="https://github.com/kakashit/rakam_client",

    author="Kaan Uğurlu",
    author_email="kugurlu94@gmail.com",

    description="Python client for rakam server.",
    long_description=open('README.rst').read(),

    packages=find_packages(),

    zip_safe=False,
    platforms='any',

    install_requires=[
        'requests>=2.4.3',
        'six>=1.10.0',
        'pytz>=2016.4',
    ],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
    ],
)



import re
import six
from pytz import utc
from datetime import datetime, date, time, timedelta, tzinfo


date_re = re.compile(
    r'(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})$'
)

time_re = re.compile(
    r'(?P<hour>\d{1,2}):(?P<minute>\d{1,2})'
    r'(?::(?P<second>\d{1,2})(?:\.(?P<microsecond>\d{1,6})\d{0,6})?)?'
)

datetime_re = re.compile(
    r'(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})'
    r'[T ](?P<hour>\d{1,2}):(?P<minute>\d{1,2})'
    r'(?::(?P<second>\d{1,2})(?:\.(?P<microsecond>\d{1,6})\d{0,6})?)?'
    r'(?P<tzinfo>Z|[+-]\d{2}:?\d{2})?$'
)


class FormatError(Exception):
    """
        Raised when date input is not well formatted.
    """
    pass


class FixedOffset(tzinfo):
    "Fixed offset in minutes east from UTC."
    def __init__(self, offset):
        if isinstance(offset, timedelta):
            self.offset = offset
            offset = self.offset.seconds // 60
        else:
            self.offset = timedelta(minutes=offset)

        self.name = "%s%02d%02d" % (
            '-' if offset < 0 else '+',
            abs(offset) / 60.,
            abs(offset) % 60
        )

    def __repr__(self):
        return self.name

    def __getinitargs__(self):
        return self.offset,

    def utcoffset(self, dt):
        return self.offset

    def tzname(self, dt):
        return self.name

    def dst(self, dt):
        return timedelta(0)


def parse_date(value):
    """
        Parses a string and return a datetime.date.
        Raises ValueError if the input is well formatted but not a valid date.
        Raises FormatError if input is not well formatted.
    """
    match = date_re.match(value)
    if match:
        kw = dict((k, int(v)) for k, v in six.iteritems(match.groupdict()))
        return date(**kw)
    else:
        raise FormatError("Not well formatted: %s" % (value,))


def parse_time(value):
    """
    Parses a string and return a datetime.time.

    This function doesn't support time zone offsets.

    Raises ValueError if the input is well formatted but not a valid time.
    Raises FormatError if the input isn't well formatted, in particular if it
    contains an offset.
    """
    match = time_re.match(value)
    if match:
        kw = match.groupdict()
        if kw['microsecond']:
            kw['microsecond'] = kw['microsecond'].ljust(6, '0')
        kw = dict((k, int(v)) for k, v in six.iteritems(kw) if v is not None)
        return time(**kw)
    else:
        raise FormatError("Not well formatted: %s" % (value,))


def parse_datetime(value):
    """
    Parses a string and return a datetime.datetime.

    This function supports time zone offsets. When the input contains one,
    the output uses an instance of FixedOffset as tzinfo.

    Raises ValueError if the input is well formatted but not a valid datetime.
    Raises FormatError if the input isn't well formatted.
    """
    match = datetime_re.match(value)
    if match:
        kw = match.groupdict()
        if kw['microsecond']:
            kw['microsecond'] = kw['microsecond'].ljust(6, '0')
        tzinfo = kw.pop('tzinfo')
        if tzinfo == 'Z':
            tzinfo = utc
        elif tzinfo is not None:
            offset = 60 * int(tzinfo[1:3]) + int(tzinfo[-2:])
            if tzinfo[0] == '-':
                offset = -offset
            tzinfo = FixedOffset(offset)
        kw = dict((k, int(v)) for k, v in six.iteritems(kw) if v is not None)
        kw['tzinfo'] = tzinfo
        return datetime(**kw)
    else:
        raise FormatError("Not well formatted: %s" % (value,))
